@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 #
 # To-do list manager.
-# Copyright (C) 2006-2009 MiKael NAVARRO
+# Copyright (C) 2006-2010 MiKael NAVARRO
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
 #
 
 """yaGTD
-Copyright (C) 2006-2009 MiKael NAVARRO
+Copyright (C) 2006-2010 MiKael NAVARRO
 
 A primitive Getting Things Done to-do list manager.
 Cmd version: line oriented command interpreters.
@@ -29,8 +29,8 @@ Cmd version: line oriented command interpreters.
 
 # Specific variables for pydoc
 __author__ = "MiKael Navarro <klnavarro@gmail.com>"
-__date__ = "November 2009"
-__version__ = "0.3.0"
+__date__ = "April 2010"
+__version__ = "0.3.2"
 
 # Include directives
 import os
@@ -55,6 +55,9 @@ TODO_TMP    = TODO_DIR + "/todo.tmp"
 REPORT_TXT  = TODO_DIR + "/report.txt"
 DONE_TXT    = TODO_DIR + "/done.txt"
 
+FORMATTED_DISPLAY = True  # show task details or no?
+USE_SHORTCUTS = False  # use all shortcuts or no?
+
 # Colorization
 COLOR_CODES = ( { 'none': "",
                   'default': "\033[0m",
@@ -105,6 +108,7 @@ END_COLOR        = COLOR_CODES['green']
 # Regexps for parsing a task line
 CONTEXT_CHAR   = "@"
 PROJECT_CHAR   = "p:"
+PROJECT_ALT    = "+"
 STATUS_CHAR    = "!"
 REFERENCE_CHAR = "ref:"
 URGENCY_CHAR    = "U:"
@@ -123,7 +127,8 @@ TIMEDELTA_MATCH = r"(\d+)([WDHM])"
 DATE_MATCH      = r"(\d\d\d\d-\d\d-\d\d)"
 
 CONTEXT_REGEXP   = re.compile(CONTEXT_CHAR + WORD_MATCH, re.IGNORECASE)
-PROJECT_REGEXP   = re.compile(PROJECT_CHAR + WORD_MATCH, re.IGNORECASE)
+PROJECT_REGEXP   = re.compile("(?:" + PROJECT_CHAR 
+                              + "|" + "\\" + PROJECT_ALT + ")" + WORD_MATCH, re.IGNORECASE)
 STATUS_REGEXP    = re.compile(STATUS_CHAR + WORD_MATCH, re.IGNORECASE)
 REFERENCE_REGEXP = re.compile(REFERENCE_CHAR + WORD_MATCH, re.IGNORECASE)
 URGENCY_REGEXP   = re.compile(URGENCY_CHAR + DIGIT_MATCH, re.IGNORECASE)
@@ -169,7 +174,7 @@ class GTD(cmd.Cmd):
         
         # Parse command args
         n = re.search(r"^(?x)(\d+)", args)  # num
-        s = re.search(r" +([@!:\-\w ]+)$", args)  # string
+        s = re.search(r" +([@!:\+\-\w ]+)$", args)  # string
 
         if n: num = int(n.group(1))
         if s: string = s.group(1)
@@ -272,7 +277,7 @@ class GTD(cmd.Cmd):
             
         # Set the title
         t['title'] = " ".join(title.split())  # remove useless blank chars too
-        
+
         return t
             
     def _dump_line(self, task):
@@ -311,7 +316,10 @@ class GTD(cmd.Cmd):
     def _disp(self, task):
         """Display the 'id' and a summary of the 'task'."""
 
-        task_line = self._dump_line(task)
+        if FORMATTED_DISPLAY:
+            task_line = self._dump_line(task)
+        else:
+            task_line = task['title']
 
         id_str = "%3d:(%f)" % (task['id'], task.priority())
         if self.colorize:  # colorize #id
@@ -395,6 +403,10 @@ class GTD(cmd.Cmd):
             for t in tasks:
                 print self._disp(t)
 
+            # Show number of tasks as result of search if more than 10
+            if len(tasks) > 10:
+                print "%d tasks found" % len(tasks)
+
         # BB: This tiny line will be damn useful for further use...
         return tasks
 
@@ -404,6 +416,65 @@ class GTD(cmd.Cmd):
         lines = self._search(regexp, quiet=True)
         return [ line['id'] for line in lines ]
 
+    def _get_deadlines_ordered(self):
+        """Get tasks ordered by deadline, oldest first."""
+
+        tasks = {}  # due -> task
+        for task in self.todo:
+            if not task['due'] or task['complete'] == 100:
+                continue
+            if not task['due'] in tasks:
+                tasks[task['due']] = []
+            tasks[task['due']].append(task)
+
+        dues = tasks.keys()
+        dues.sort()
+
+        ordered = []
+
+        for due in dues:
+            ordered.extend(tasks[due])
+
+        return ordered
+
+    def _get_deadline_string(self, task):
+        """Return formatted string of a deadline task."""
+
+        now = datetime.datetime.now()
+        daystart = datetime.datetime(year=now.year, 
+                                     month=now.month,
+                                     day=now.day)
+        dayend = daystart + datetime.timedelta(days=1)
+
+        s = ""
+        due = task.get('due', None)
+        if due:
+            left = due - daystart
+            if due < daystart:
+                days = -left.days
+                if days == 1:
+                    s = " [overdue one day]"
+                else:
+                    s = " [overdue %s days]" % days
+            elif due < dayend:
+                s = " [due today]"
+            else:
+                if left.days == 1:
+                    s = " [one day left]"
+                else:
+                    s = " [%s days left]" % left.days
+
+        return s
+
+        s = ""
+        due = task.get('due', None)
+        if due:
+            left = due - now
+            if left < datetime.timedelta(days=1):
+                s = " [due today]"
+            elif left > datetime.timedelta(0):
+                s = " [%s days left]" % left.days
+
     #
     # Edition.
     #
@@ -416,6 +487,8 @@ class GTD(cmd.Cmd):
         if task_id: 
             print "Task #%d added" % task_id
 
+    do_a = do_add
+
     def do_del(self, id):
         """Delete given task:
         GTD> del #id"""
@@ -490,6 +563,7 @@ class GTD(cmd.Cmd):
                     print "Task #%d (%s) completed." % (idx, task['title'])
 
     do_done = do_close
+    do_do = do_close
 
     def do_replace(self, id_line):
         """Replace the entire task by a new one:
@@ -542,6 +616,8 @@ class GTD(cmd.Cmd):
                 if task_modified:
                     print "Task #%d appended" % idx
 
+    do_addto = do_append
+
     def do_appendall(self, args):  # BB
         """Add new elements to several tasks but leaves other elements unchanged:
         GTD> appendall regexp @newcontext p:NewProject"""
@@ -579,6 +655,8 @@ class GTD(cmd.Cmd):
         else:
             print "Please specify the id of the task to modify first."
 
+    do_mod = do_modify
+
     def do_modifyall(self, args):  # BB
         """Add/change elements of several tasks but leave each other unchanged:
         GTD> modifyall regexp @othercontext p:UpdtProject"""
@@ -640,6 +718,8 @@ class GTD(cmd.Cmd):
         if idx and level and re.match(DIGIT_MATCH, level):
             self.do_modify("%d U:%d" % (idx, int(level)))
 
+    do_u = do_urgency
+
     def do_importance(self, id_level):
         """Set the Importance (5=crucial,4=high,3=normal,2=low,1=someday):
         GTD> importance #id lvl"""
@@ -650,6 +730,8 @@ class GTD(cmd.Cmd):
         if idx and level and re.match(DIGIT_MATCH, level):
             self.do_modify("%d I:%d" % (idx, int(level)))
 
+    do_i = do_importance
+
     def do_complete(self, id_percent):
         """Set the percent complete:
         GTD> complete #id percentage"""
@@ -660,6 +742,8 @@ class GTD(cmd.Cmd):
         if idx and percent and re.match(NUMBER_MATCH, percent):
             self.do_modify("%d C:%d" % (idx, int(percent)))
 
+    do_c = do_complete
+
     def do_time(self, id_time):
         """Set the Time requiered to accomplish the task (in Weeks, Days, Hours or Minutes):
         GTD> time #id n[WDHM]"""
@@ -680,6 +764,9 @@ class GTD(cmd.Cmd):
         if idx and recurrence and re.match(TIMEDELTA_MATCH, recurrence):
             self.do_modify("%d R:%s" % (idx, recurrence))
 
+    do_recurs = do_recurrence
+    do_rec = do_recurrence
+
     def do_start(self, id_date):
         """Set the Start / creation date:
         GTD> start #id YYYY-MM-DD"""
@@ -700,6 +787,21 @@ class GTD(cmd.Cmd):
         if idx and date and re.match(DATE_MATCH, date):
             self.do_modify("%d D:%s" % (idx, date))
 
+    def do_duein(self, id_doff):
+        """Set the Due (target) date in +/- days:
+        GTD> duein #id +/-days"""
+
+        # Parse command line
+        idx, days_offset = self._parse_args(id_doff)
+        
+        if idx and days_offset:
+            # Compute the new date= now + days_offset
+            now = datetime.datetime.now()
+            dayend = now + datetime.timedelta(int(days_offset))
+            date = dayend.strftime("%Y-%m-%d")
+
+            self.do_modify("%d D:%s" % (idx, date))
+
     def do_end(self, id_date):
         """Set the End, closure date:
         GTD> end #id YYYY-MM-DD"""
@@ -754,6 +856,26 @@ class GTD(cmd.Cmd):
 
     do_ls = do_list
 
+    def do_listinbox(self, nb):
+        """List #nb quick added tasks (i.e. without context; ordered by #id):
+        GTD> listinbox [#nb]"""
+
+        # Parse command line
+        nb = self._parse_args(nb)[0]
+
+        tasks = [ t for t in self.todo if t['complete'] < 100 and not t['context'] ]  # without context
+        if nb:  # display only nb tasks
+            tasks = tasks[:nb]
+
+        for t in tasks:
+            print self._disp(t)
+
+        # Show number of tasks as result of search if more than 10
+        if len(tasks) > 10:
+            print "%d inbox tasks found" % len(tasks)
+
+    do_li = do_listinbox
+
     def do_listref(self, nb):
         """List #nb ref tasks (ordered by #id):
         GTD> listref [#nb]"""
@@ -770,7 +892,7 @@ class GTD(cmd.Cmd):
 
         # Show number of tasks as result of search if more than 10
         if len(tasks) > 10:
-            print "%d tasks found" % len(tasks)
+            print "%d references found" % len(tasks)
 
     do_lr = do_listref
 
@@ -787,6 +909,9 @@ class GTD(cmd.Cmd):
             if task:
                 print self._show(task)
 
+    do_view = do_show
+    do_v = do_show
+
     def do_sort(self, nb):
         """Sort #nb tasks by priority (appear only after start date):
         GTD> sort [#nb]"""
@@ -806,6 +931,10 @@ class GTD(cmd.Cmd):
         for t in tasks:
             print self._disp(t)
 
+        # Show number of tasks as result of search if more than 10
+        if len(tasks) > 10:
+            print "%d tasks found" % len(tasks)
+
     def do_today(self, nb):
         """Sort #nb tasks by priority, and shows only today
         (the time now before and after start date):
@@ -829,6 +958,12 @@ class GTD(cmd.Cmd):
                     print self._disp(t)
             else:
                 print self._disp(t) 
+
+        # Show number of tasks as result of search if more than 10
+        if len(tasks) > 10:
+            print "%d today tasks found" % len(tasks)
+
+    do_now = do_today
             
     def do_listpri(self, regexp):
         """Sort tasks matching given regexp (appear only after start date):
@@ -846,8 +981,13 @@ class GTD(cmd.Cmd):
 
         for t in tasks:
             print self._disp(t)
-
     
+        # Show number of tasks as result of search if more than 10
+        if len(tasks) > 10:
+            print "%d tasks found" % len(tasks)
+
+    do_lp = do_listpri
+
     def do_order(self, nb_attr):
         """Order #nb tasks by context/project/status/reference and priority
         (appear only after start date):
@@ -878,6 +1018,12 @@ class GTD(cmd.Cmd):
             for t in tasks:
                 print self._disp(t)
 
+            # Show number of tasks as result of search if more than 10
+            if len(tasks) > 10:
+                print "%d tasks found" % len(tasks)
+
+    do_o = do_order
+
     def do_status(self, line):
         """Display projects statuses (percent complete):
         GTD> status"""
@@ -903,6 +1049,95 @@ class GTD(cmd.Cmd):
                             
     do_summary = do_status
 
+    def do_contexts(self, line):
+        """Display contexts and next task for each context:
+        GTD> contexts"""
+
+        for context, tasks in self.todo.order('context').items():
+            percent = 0  # compute project's percent complete
+            count = 0
+            next = None
+            for t in tasks:
+                print t['title'], t.priority(), t['complete']
+                percent += int(t['complete'])
+                if t['complete'] != 100:
+                    print "->", t['title']
+                    count += 1
+                if not next and t['status'] and 'Next' in t['status'] and t['complete'] != 100: 
+                    next = t
+
+            percent /= len(tasks) 
+            if percent == 100: continue 
+
+            # Section title 
+            print self._colorize(CONTEXT_CHAR + context) + ':', 
+            print "%d" % count,
+            if next:
+                projects = next['project']
+                if projects:
+                    print self._colorize(PROJECT_CHAR + projects[0]),
+                print self._colorize("!Next: %s" % next['title'])
+            else:
+                print
+ 
+    def do_deadlines(self, line):
+        """Display upcoming tasks ordered by deadline:
+        GTD> deadlines #nb"""
+
+        # Parse command line
+        nb = self._parse_args(line)[0]
+ 
+        ordered = self._get_deadlines_ordered()
+
+        # Keep only upcoming tasks
+        now = datetime.datetime.now()
+        daystart = datetime.datetime(year=now.year, 
+                                     month=now.month,
+                                     day=now.day)
+        dayend = daystart + datetime.timedelta(days=1)
+        ordered = [t for t in ordered if t.get('due', None) >= daystart]
+
+        if nb: ordered = ordered[:nb]  # display only nb tasks
+
+        for t in ordered:
+            # note about due
+            s = self._get_deadline_string(t)
+
+            print self._disp(t) + s
+
+        # Show number of tasks as result of search if more than 10
+        if len(ordered) > 10:
+            print "%d due tasks found" % len(ordered)
+
+    def do_overdue(self, line):
+        """Display overdue tasks ordered by deadline (including due today):
+        GTD> overdue #nb"""
+
+        # Parse command line
+        nb = self._parse_args(line)[0]
+
+        ordered = self._get_deadlines_ordered()
+
+        # Keep only overdue tasks
+        now = datetime.datetime.now()
+        daystart = datetime.datetime(year=now.year, 
+                                     month=now.month,
+                                     day=now.day)
+        dayend = daystart + datetime.timedelta(days=1)
+        ordered = [t for t in ordered if t.get('due', None) < dayend]
+    
+        if nb: ordered = ordered[:nb]  # display only nb tasks
+
+        for t in ordered:
+            # note about due
+            s = self._get_deadline_string(t)
+
+            print self._disp(t) + s
+
+        # Show number of tasks as result of search if more than 10
+        if len(ordered) > 10:
+            print "%d overdue tasks found" % len(ordered)
+
     #
     # Select/monitor.
     #
@@ -941,6 +1176,10 @@ class GTD(cmd.Cmd):
         for t in (self.todo.find("id", idx) for idx in self.tasks_selected):
             print self._disp(t)
 
+        # Show number of tasks as result of search if more than 10
+        if len(tasks) > 10:
+            print "%d selected tasks found" % len(tasks)
+
     #
     # Re-search.
     #
@@ -951,12 +1190,16 @@ class GTD(cmd.Cmd):
 
         self._search(regexp)
 
+    do_s = do_search
+
     def do_searchall(self, regexp): 
         """Retrieve tasks matching given regexp, including completed ones:
         GTD> searchall regexp""" 
         
         self._search(regexp, completed=True)
         
+    do_sa = do_searchall
+
     #
     # IO functions.
     #
@@ -965,8 +1208,9 @@ class GTD(cmd.Cmd):
         """Load from a todotxt file:
         GTD> load path/to/todo.txt"""
 
+        self.todotxt = todotxt  # ok, save file path
         try:
-            f = open(todotxt, 'r')
+            f = open(self.todotxt, 'r')
             try:
                 self.todo.erase()  # clean
                 nb = 0
@@ -977,12 +1221,13 @@ class GTD(cmd.Cmd):
                     nb += 1
             finally:
                 f.close()
-                self.todotxt = todotxt  # ok, save file path
                 print "%d tasks loaded from '%s'" % (nb, todotxt)
         except IOError, err:
             # 'file not found' exception?
             if err.errno != errno.ENOENT: raise  # raise all others
-            print err  # and continue
+            print err  # and continue (new file created)
+
+    do_l = do_load
 
     def do_save(self, todotxt):
         """Save to a todotxt file:
@@ -1000,6 +1245,9 @@ class GTD(cmd.Cmd):
         except IOError, err:
             print err  # and continue
 
+    do_write = do_save
+    do_w = do_save
+
     def do_archive(self, donetxt):
         """Archive completed tasks:
         GTD> archive [path/to/done.txt]"""
@@ -1014,8 +1262,7 @@ class GTD(cmd.Cmd):
             try:
                 for t in done_lst:
                     f.write(self._dump_line(t) + "\n")
-                # If all okey, remove done/close tasks from current list
-                self.todo = [ t for t in self.todo if t['complete'] < 100 ]
+                    self.todo.supp(t['id'])  # remove done/close tasks from current list
             finally:
                 f.close()
                 print "%d tasks archived as '%s'" % (len(done_lst), donetxt)
@@ -1055,6 +1302,8 @@ class GTD(cmd.Cmd):
                                'char': REFERENCE_CHAR },
                              )
 
+                unprinted = self.todo[:]  # to detect unprinted tasks
+
                 for attr in prt_list:
                     print attr['title'].capitalize()
                     print "=" * len(attr['title'])
@@ -1062,17 +1311,33 @@ class GTD(cmd.Cmd):
 
                     for a, ts in self.todo.order(attr['name']).items():
                         attr_name = attr['char'] + a  # attr (full) name
-                        print attr_name.lower()
+                        print attr_name
                         print "-" * len(attr_name)
                         print
                         
                         tasks = [ t for t in ts if t['complete'] < 100 ] 
                         for t in tasks:  # task details
                             print "-", t['title']
+                            try:  # ok, remove from unprinted list
+                                unprinted.remove(t)
+                            except ValueError:
+                                pass
                         print
                     else:
-                        print ".."
+                        #print ".."
                         print
+
+                # No context, project, status and reference tasks
+                unprinted = [t for t in unprinted if t['complete'] < 100]
+                if unprinted:  # if some tasks left
+                    print "Otherwise Unmentioned Open Tasks"
+                    print "================================"
+                    print
+
+                    for t in unprinted:
+                        print "-", t['title']
+                    print
+
             finally:
                 sys.stdout = sys.__stdout__  # restore stdout
                 f.close()
@@ -1120,6 +1385,9 @@ Type 'help' or '?' for more commands/options."""
         self.do_save("")  # auto save (default location)
         sys.exit()
 
+    do_quit = do_EOF
+    do_exit = do_EOF
+    do_q = do_EOF
 
 #
 # Main entry point.
@@ -1135,7 +1403,7 @@ def main(options, todotxt=TODO_TXT, todoyaml=None):
     elif todoyaml:
         gtd_cmd.do_import(todoyaml)
 
-    if options.color:
+    if options and options.color:
         gtd_cmd.colorize = True
 
     try:
